﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace Parallel
{
	public static class ParallelExtensions
	{
		private static readonly int processors = Environment.ProcessorCount;
		public static IParallelEnumerable<T> ToParallel<T>(this IEnumerable<T> enumerable)
		{
			return new ParallelContainer<T>(enumerable);
		}

		public static IEnumerable<T> ToEnumerable<T>(this IParallelEnumerable<T> enumerable)
		{
			return (IEnumerable<T>)enumerable;
		}

		public static IParallelEnumerable<T> Where<T>(this IParallelEnumerable<T> parallelEnumerable, Func<T, bool> predicate)
		{
			return Execute<T, T>(parallelEnumerable, processors, parameterEnumerable => parameterEnumerable.ToEnumerable().Where(predicate).ToParallel());
		}

		public static IParallelEnumerable<S> Select<T, S>(this IParallelEnumerable<T> parallelEnumerable, Func<T, S> project)
		{
		    return Execute<T, S>(parallelEnumerable, processors, parameterEnumerable => parameterEnumerable.ToEnumerable().Select(project).ToParallel());
		}

		private static IEnumerable<IParallelEnumerable<T>> Partitions<T>(IParallelEnumerable<T> data, int processors)
		{
			int itemsPerPartition = ItemsPerPartition<T>(data, processors);
			while(data.Any())
			{
				yield return data.Take(itemsPerPartition).ToParallel();
				data = data.Skip(itemsPerPartition).ToParallel();
			}
		}

		private static int ItemsPerPartition<T>(IParallelEnumerable<T> data, int processors)
		{
			return (int)Math.Ceiling((float)data.Count() / processors);
		}

		private static IParallelEnumerable<TResult> Execute<T, TResult>(IParallelEnumerable<T> data, int processors, Func<IParallelEnumerable<T>, IParallelEnumerable<TResult>> func)
		{
			if(data.Count() <= 0) return new ParallelContainer<TResult>();

			var partitions = Partitions(data, processors).ToList();
			using(CountDown countDown = new CountDown(partitions.Count()))
			{
				var listOfExecutors = new List<Executor<T, TResult>>();
				foreach(var partition in partitions)
				{
					var executor = new Executor<T, TResult>(partition, func);
					listOfExecutors.Add(executor);
					ThreadPool.QueueUserWorkItem((object state) =>
					{
						((Executor<T, TResult>)state).Execute();
						countDown.Signal();
					}, executor);
				}

				countDown.Wait();
				List<TResult> result = new List<TResult>();
				foreach(var executor in listOfExecutors)
					result.AddRange(executor.Value);

				return new ParallelContainer<TResult>(result);
			}
		}
	}
}
